@@ -22,6 +22,7 @@ all_models = [
     'ComponentAnalysis',
     'DatepartRegression',
     "UnivariateRegression",
+    "Greykite",
 ]
 default = [
     'ZeroesNaive',
@@ -49,7 +50,7 @@ superfast = [
     'GLS',
     'SeasonalNaive',
 ]
-# relatively fast at any scale
+# relatively fast
 fast = [
     'ZeroesNaive',
     'LastValueNaive',
@@ -57,12 +58,14 @@ fast = [
     'GLS',
     'GLM',
     'ETS',
-    'WindowRegression',
-    'GluonTS',
+    # 'WindowRegression',
+    # 'GluonTS',
     'VAR',
     'SeasonalNaive',
     'VECM',
     'ComponentAnalysis',
+    'DatepartRegression',
+    'UnivariateRegression',
 ]
 # models that can scale well if many CPU cores are available
 parallel = [
@@ -72,6 +75,7 @@ parallel = [
     'GLM',
     'UnobservedComponents',
     'MotifSimulation',
+    "Greykite",
 ]
 # models that should be fast given many CPU cores
 fast_parallel = list(set(parallel + fast))
@@ -125,10 +129,11 @@ recombination_approved = [
     'GluonTS',
     'RollingRegression',
     'VAR',
-    'WindowRegression',
+    # 'WindowRegression',
     'TensorflowSTS',
     'TFPRegression',
     'UnivariateRegression',
+    "Greykite",
 ]
 # USED IN AUTO_MODEL for models that don't share information among series
 no_shared = [
@@ -143,7 +148,8 @@ no_shared = [
     'UnobservedComponents',
     'TensorflowSTS',
     "GLS",
-    "UnivariateRegression"
+    "UnivariateRegression",
+    "Greykite",
     # 'MotifSimulation',
     # 'DatepartRegression',
 ]
@@ -159,6 +165,7 @@ regressor = [
     'WindowRegression',
     'VAR',
     'DatepartRegression',
+    "Greykite",
     # "UnivariateRegression",
 ]
 no_shared_fast = list(set(no_shared).intersection(set(fast_parallel)))
@@ -180,3 +187,7 @@ model_lists = {
     "gpu": gpu,
     "regressor": regressor,
 }
+
+
+def auto_model_list(n_jobs, n_series, frequency):
+    pass
